#pragma once
#include <cfloat>
#include <climits>
#include <thread>
#include <numeric>
#include <iterator>
#include <vector>

/**
 * @brief Computes an alternating sum of the elements in the array arr,
 * that is arr[0] - arr[1] + arr[2] - ...
 * @param arr - the array of integers
 * @param num_threads the number of threads to be used for computation
 */
int AltSumParallel(const std::vector<int> &arr, size_t num_threads)
{
    std::vector<int> results(num_threads, 0);
    std::vector<std::thread> threads(num_threads);

    size_t chunk_size = arr.size() / num_threads;
    size_t start = 0;
    for (size_t i = 0; i < num_threads; i++)
    {
        size_t end = (i == num_threads - 1) ? arr.size() : start + chunk_size;
        threads[i] = std::thread([start, end, &arr, &results, i]
                                 {
            for (size_t j = start; j < end; j++) {
                results[i] += (j % 2 == 0) ? arr[j] : -arr[j];
            } });
        start = end;
    }

    for (size_t i = 0; i < num_threads; i++)
    {
        threads[i].join();
    }

    return std::accumulate(results.begin(), results.end(), 0);
}

//-----------------------------------------------------------------------------
